local _, addon = ...

local MapCanvasMixin = { }

local function OnPinReleased(pinPool, pin)
	FramePool_HideAndClearAnchors(pinPool, pin);
	pin:OnReleased();

	pin.pinTemplate = nil;
	pin.owningMap = nil;
end

local function OnPinMouseUp(pin, button, upInside)
	pin:OnMouseUp(button);
	if upInside then
		pin:OnClick(button);
	end
end

function MapCanvasMixin:RemovePin(pin)
	if pin:GetNudgeSourceRadius() > 0 then
		self.pinNudgingDirty = true;
	end

	self.pinPools[pin.pinTemplate]:Release(pin);
	--self.target.ScrollContainer:MarkCanvasDirty();
end


local MapCanvasProxyMixin = {
	isProxy = true
}

-- pass thru
function MapCanvasProxyMixin:AcquirePin(pinTemplate, ...)
	if not self.pinPools[pinTemplate] then
		local pinTemplateType = self.pinTemplateTypes[pinTemplate] or "FRAME";
		self.pinPools[pinTemplate] = CreateFramePool(pinTemplateType, self.target:GetCanvas(), pinTemplate, OnPinReleased);
	end

	local pin, newPin = self.pinPools[pinTemplate]:Acquire();

	if newPin then
		local isMouseClickEnabled = pin:IsMouseClickEnabled();
		local isMouseMotionEnabled = pin:IsMouseMotionEnabled();

		if isMouseClickEnabled then
			pin:SetScript("OnMouseUp", OnPinMouseUp);
			pin:SetScript("OnMouseDown", pin.OnMouseDown);
		end

		if isMouseMotionEnabled then
			if newPin then
				-- These will never be called, just define a OnMouseEnter and OnMouseLeave on the pin mixin and it'll be called when appropriate
				assert(pin:GetScript("OnEnter") == nil);
				assert(pin:GetScript("OnLeave") == nil);
			end
			pin:SetScript("OnEnter", pin.OnMouseEnter);
			pin:SetScript("OnLeave", pin.OnMouseLeave);
		end

		pin:SetMouseClickEnabled(isMouseClickEnabled);
		pin:SetMouseMotionEnabled(isMouseMotionEnabled);
	end

	pin.pinTemplate = pinTemplate;
	pin.owningMap = self;

	--if newPin then
	--	pin:OnLoad();
	--end

	--self.target.ScrollContainer:MarkCanvasDirty();
	pin:Show();
	pin:OnAcquired(...);

	return pin;
end

function MapCanvasProxyMixin:AddDataProvider(dataProvider)
	self.dataProviders[dataProvider] = true;
	dataProvider:OnAdded(self);
end

-- copied
function MapCanvasProxyMixin:ApplyPinPosition(pin, normalizedX, normalizedY, insetIndex)
	if insetIndex then
		if self.mapInsetsByIndex and self.mapInsetsByIndex[insetIndex] then
			self.mapInsetsByIndex[insetIndex]:SetLocalPinPosition(pin, normalizedX, normalizedY);
			pin:ApplyFrameLevel();
		end
	else
		pin:ClearAllPoints();
		if normalizedX and normalizedY then
			local x = normalizedX;
			local y = normalizedY;

			local nudgeVectorX, nudgeVectorY = pin:GetNudgeVector();
			if nudgeVectorX and nudgeVectorY then
				local finalNudgeFactor = pin:GetNudgeFactor() * pin:GetNudgeTargetFactor() * pin:GetNudgeZoomFactor();
				x = normalizedX + nudgeVectorX * finalNudgeFactor;
				y = normalizedY + nudgeVectorY * finalNudgeFactor;
			end

			local canvas = self.target:GetCanvas(); -- added target
			local scale = pin:GetScale();
			pin:SetParent(canvas);
			pin:ApplyFrameLevel();
			pin:SetPoint("CENTER", canvas, "TOPLEFT", (canvas:GetWidth() * x) / scale, -(canvas:GetHeight() * y) / scale);
		end
	end
end

-- pass thru
function MapCanvasProxyMixin:GetGlobalPinScale()
	return self.target:GetGlobalPinScale()
end

function MapCanvasProxyMixin:GetMapID()
	return self.target:GetMapID()
end

-- pass thru
function MapCanvasProxyMixin:GetPinFrameLevelsManager()
	return self.target:GetPinFrameLevelsManager()
end

function MapCanvasProxyMixin:IsShown()
	return self.target:IsShown()
end

-- pass thru
function MapCanvasProxyMixin:RemovePin(...)
	MapCanvasMixin.RemovePin(self,...)
end

-- copied
function MapCanvasProxyMixin:SetPinPosition(pin, normalizedX, normalizedY, insetIndex)
	self:ApplyPinPosition(pin, normalizedX, normalizedY, insetIndex);
	if not pin:IgnoresNudging() then
		if pin:GetNudgeSourceRadius() > 0 then
			-- If we nudge other things we need to recalculate all nudging.
			self.pinNudgingDirty = true;
		else
			self.pinsToNudge[#self.pinsToNudge + 1] = pin;
		end
	end
end

-- copied
function MapCanvasProxyMixin:OnCanvasScaleChanged()
	--self:AdjustDetailLayerAlpha();

	--if self.mapInsetsByIndex then
	--	for insetIndex, mapInset in pairs(self.mapInsetsByIndex) do
	--		mapInset:OnCanvasScaleChanged();
	--	end
	--end

	self:CallMethodOnPinsAndDataProviders("OnCanvasScaleChanged");
end

-- copied
function MapCanvasProxyMixin:CallMethodOnPinsAndDataProviders(methodName, ...)
	for dataProvider in pairs(self.dataProviders) do
		dataProvider[methodName](dataProvider, ...);
	end

	for pin in self:EnumerateAllPins() do
		pin[methodName](pin, ...);
	end
end

-- copied
function MapCanvasProxyMixin:EnumerateAllPins()
	local currentPoolKey, currentPool = next(self.pinPools, nil);
	local currentPin = nil;
	return function()
		if currentPool then
			currentPin = currentPool:GetNextActive(currentPin);
			while not currentPin do
				currentPoolKey, currentPool = next(self.pinPools, currentPoolKey);
				if currentPool then
					currentPin = currentPool:GetNextActive();
				else
					break;
				end
			end
		end

		return currentPin;
	end, nil;
end



local function init(proxy)
	proxy.dataProviders = { }
	proxy.pinPools = { }
	proxy.pinTemplateTypes = { }
	hooksecurefunc(proxy.target, "OnMapChanged", function()
		for dataProvider in pairs(proxy.dataProviders) do
			dataProvider:OnMapChanged();
		end
	end)
	hooksecurefunc(proxy.target, "OnCanvasScaleChanged", function()
		proxy:OnCanvasScaleChanged()
	end)
end

local function OnEvent(event, arg1)
	if (event == "ADDON_LOADED") then
		addon.AddProxy(WorldMapFrame, MapCanvasProxyMixin, init)
		addon.AddProxy(BattlefieldMapFrame, MapCanvasProxyMixin, init)
	end
end

addon.RegisterEvent("ADDON_LOADED", OnEvent)
