local _, addon = ...

local C_Map = C_Map
local C_VignetteInfo = C_VignetteInfo
local CreateFrame = CreateFrame
local CreateFromMixins = CreateFromMixins
local WorldMapFrame
local BattlefieldMapFrame

local TomCatsVignetteTooltip = TomCatsVignetteTooltip

local maps = { }
local atlasSizes = { }
local vignettes
local interval, minInterval, maxInterval = 1, 1/15, 1
local timeSinceLastUpdate = 0
local vignetteInfoCache = { }

local PIN_STATUS = {
	STATIC = {
		texture = "VignetteKill",
		scaleFactor = 1.00
	},
	SPAWNED = {
		texture = "nazjatar-nagaevent",
		scaleFactor = 1.40
	},
	COMPLETE = {
		texture = "Capacitance-General-WorkOrderCheckmark",
		scaleFactor = 0.90
	}
}

local function ShowHide(element, condition)
	if (condition) then
		element:Show()
		return true
	end
	element:Hide()
	return false
end

local function rescale(pin)
	local atlasName = pin.Texture:GetAtlas()
	if (atlasName) then
		if (not atlasSizes[atlasName]) then
			atlasSizes[atlasName] = { pin.Texture:GetSize() }
		end
		local zoom = pin:GetParent():GetScale()
		local mapData = maps[pin:GetMap()]
		local sizeX = atlasSizes[atlasName][1] / zoom * (pin.scaleFactor or 1) * mapData.iconScale
		local sizeY = atlasSizes[atlasName][2] / zoom * (pin.scaleFactor or 1) * mapData.iconScale
		pin.Texture:SetSize(sizeX, sizeY)
		pin.HighlightTexture:SetSize(sizeX, sizeY)
		pin:SetSize(sizeX, sizeY)
	end
end

local function setPinIcon(pin, status)
	pin.Texture:SetAtlas(status.texture, true)
	pin.HighlightTexture:SetAtlas(status.texture, true)
	pin.scaleFactor = status.scaleFactor
	rescale(pin)
end

local function updateVignettePin(pin)
	local vignette = pin.vignette
	if (vignette) then
		if (pin.isSpawned) then
			setPinIcon(pin, PIN_STATUS.SPAWNED)
		else
			local isCompleted = vignette.isCompleted
			if (isCompleted) then
				setPinIcon(pin, PIN_STATUS.COMPLETE)
			else
				setPinIcon(pin, PIN_STATUS.STATIC)
			end
		end
	end
end

local function OnUpdate(_, elapsed)
	timeSinceLastUpdate = timeSinceLastUpdate + elapsed
	if (timeSinceLastUpdate > interval) then
		timeSinceLastUpdate = 0
		local mapID = C_Map.GetBestMapForUnit("player")
		local affectedMaps
		for mapFrame, mapData in pairs(maps) do
			if (next(mapData.pins) and mapID == mapFrame:GetMapID() and mapFrame:IsShown()) then
				affectedMaps = affectedMaps or { }
				table.insert(affectedMaps, mapData)
			end
		end
		local newInterval
		if (affectedMaps) then
			vignettes = C_VignetteInfo.GetVignettes()
			local vignetteGUIDsByVignetteID
			if (#vignettes ~= 0) then
				newInterval = minInterval
				for _, v in ipairs(vignettes) do
					local vignetteID = vignetteInfoCache[v]
					if (not vignetteID) then
						local vignetteInfo = C_VignetteInfo.GetVignetteInfo(v)
						vignetteID = vignetteInfo.vignetteID
						vignetteInfoCache[v] = vignetteID
					end
					vignetteGUIDsByVignetteID = vignetteGUIDsByVignetteID or { }
					vignetteGUIDsByVignetteID[vignetteID] = vignetteGUIDsByVignetteID[vignetteID] or { }
					table.insert(vignetteGUIDsByVignetteID[vignetteID], v)
				end
			end
			for vignetteID in pairs(affectedMaps[1].pins) do
				local vignetteGUIDs = vignetteGUIDsByVignetteID and vignetteGUIDsByVignetteID[vignetteID]
				for _, mapInfo in ipairs(affectedMaps) do
					local pin = mapInfo.pins[vignetteID]
					local spawned
					if (pin) then
						if (vignetteGUIDs) then
							local coordinates
							if (not coordinates and coordinates ~= false) then
								local idx = C_VignetteInfo.FindBestUniqueVignette(vignetteGUIDs)
								if (idx) then
									coordinates = C_VignetteInfo.GetVignettePosition(vignetteGUIDs[idx], mapID)
								else
									coordinates = false
								end
							end
							if (coordinates) then
								pin.isSpawned = true
								updateVignettePin(pin)
								pin:SetPosition(coordinates:GetXY())
								spawned = true
							end
						end
						if (not spawned) then
							pin.isSpawned = false
							updateVignettePin(pin)
							pin:SetPosition(pin.vignette:GetLocation())
						end
					end
				end
			end
		end
		interval = newInterval or maxInterval
	end
end

local TomCatsMapCanvasDataProviderMixin = { } --CreateFromMixins(MapCanvasDataProviderMixin)

-- copied
function TomCatsMapCanvasDataProviderMixin:GetMap()
	return self.owningMap
end

function TomCatsMapCanvasDataProviderMixin:OnAdded(owningMap)
	self.owningMap = owningMap
	--owningMap:GetPinFrameLevelsManager():
	--InsertFrameLevelAbove("PIN_FRAME_LEVEL_TOMCATS", "PIN_FRAME_LEVEL_AREA_POI", 200)
end

function TomCatsMapCanvasDataProviderMixin:RefreshAllData()
	local mapFrame = self:GetMap()
	local mapData = maps[mapFrame]
	if (mapData) then
		local vignetteInfo = addon.getVignettes(mapFrame:GetMapID())
		if (vignetteInfo) then
			for vignetteID, vignette in pairs(vignetteInfo) do
				if (vignette.isVisible) then
					if (not mapData.pins[vignetteID]) then
						ShowHide(mapFrame:AcquirePin("TomCatsMapPinTemplate", vignette), not self.hideAll);
					end
				end
			end
		end
	end
end

function TomCatsMapCanvasDataProviderMixin:OnCanvasScaleChanged()
	local mapData = maps[self:GetMap()]
	if (mapData) then
		for _, pin in pairs(mapData.pins) do
			rescale(pin)
		end
	end
end

function TomCatsMapCanvasDataProviderMixin:OnMapChanged()
	local mapFrame = self:GetMap()
	local mapData = maps[mapFrame]
	if (mapData) then
		for vignetteID, pin in pairs(mapData.pins) do
			mapData.pins[vignetteID] = nil
			pin:Hide()
			mapFrame:RemovePin(pin)
		end
	end
	self:RefreshAllData()
end

local eventFrame = CreateFrame("Frame")

local function setQuestFocus(self, questID)
	local mapFrame = self:GetMap()
	local mapData = maps[mapFrame]
	self.hideAll = questID ~= nil
	for _, pin in pairs(mapData.pins) do
		ShowHide(pin, not self.hideAll)
	end
end

local function setupMapProvider(map, iconScale)
	assert(map)
	if (not maps[map]) then
		maps[map] = {
			pins = { },
			iconScale = iconScale,
			mapFrame = map
		}
		local provider = CreateFromMixins(TomCatsMapCanvasDataProviderMixin)
		map:AddDataProvider(provider)
		table.insert(addon.ruleListeners, function()
			provider:OnMapChanged()
		end)
		if (map == WorldMapFrame) then
			--hooksecurefunc(WorldMapFrame,"SetFocusedQuestID", setQuestFocus)
			--hooksecurefunc(WorldMapFrame,"ClearFocusedQuestID", setQuestFocus)
		end
	end
end

local function OnEvent(event, arg1)
	if (event == "ADDON_LOADED") then
		if (not WorldMapFrame and _G["WorldMapFrame"]) then
			WorldMapFrame = addon.GetProxy(_G["WorldMapFrame"])
			setupMapProvider(WorldMapFrame, 0.7)
		end
		if (not BattlefieldMapFrame and _G["BattlefieldMapFrame"]) then
			BattlefieldMapFrame = addon.GetProxy(_G["BattlefieldMapFrame"])
			setupMapProvider(BattlefieldMapFrame, 0.7)
		end
	end
end

addon.RegisterEvent("ADDON_LOADED", OnEvent)
eventFrame:SetScript("OnUpdate", OnUpdate)

TomCatsMapCanvasPinMixin = { } --CreateFromMixins(MapCanvasPinMixin)

-- copied
function TomCatsMapCanvasPinMixin:ApplyFrameLevel()
	local frameLevel = self:GetMap():GetPinFrameLevelsManager():GetValidFrameLevel("PIN_FRAME_LEVEL_AREA_POI");
	self:SetFrameLevel(frameLevel + 1);
	--self:SetFrameLevel(2000);
end

-- copied
function TomCatsMapCanvasPinMixin:GetMap()
	return self.owningMap;
end

-- copied
function TomCatsMapCanvasPinMixin:GetNudgeSourceRadius()
	return self.nudgeSourceRadius or 0;
end

-- copied
function TomCatsMapCanvasPinMixin:GetNudgeTargetFactor()
	return self.nudgeTargetFactor or 0;
end

-- copied
function TomCatsMapCanvasPinMixin:GetNudgeVector()
	return self.nudgeVectorX, self.nudgeVectorY;
end

-- copied
function TomCatsMapCanvasPinMixin:IgnoresNudging()
	return self.insetIndex or (self:GetNudgeSourceRadius() == 0 and self:GetNudgeTargetFactor() == 0);
end


function TomCatsMapCanvasPinMixin:OnAcquired(vignette)
	self.vignette = vignette
	local mapData = maps[self:GetMap()]
	mapData.pins[vignette.ID] = self
	self:SetPosition(vignette:GetLocation())
	updateVignettePin(self)
	self:UseFrameLevelType("PIN_FRAME_LEVEL_TOMCATS")
	rescale(self)
	self:ApplyCurrentScale()
end

function TomCatsMapCanvasPinMixin:OnReleased()
	if (self.vignette) then
		maps[self:GetMap()].pins[self.vignette.ID] = nil
	end
	self.vignette = nil
	self.update = nil
	self.isSpawned = nil
	self.scaleFactor = nil
	self.Texture:SetAtlas(nil)
	self.HighlightTexture:SetAtlas(nil)
end

function TomCatsMapCanvasPinMixin:OnMouseEnter()
	TomCatsVignetteTooltip:SetOwner(self)
end

function TomCatsMapCanvasPinMixin:OnMouseLeave()
	TomCatsVignetteTooltip:SetOwner()
end

function TomCatsMapCanvasPinMixin:OnUpdate()
	if (self.update) then self:update() end
end

-- copied
function TomCatsMapCanvasPinMixin:SetPosition(normalizedX, normalizedY, insetIndex)
	self.normalizedX = normalizedX;
	self.normalizedY = normalizedY;
	self.insetIndex = insetIndex;
	self:GetMap():SetPinPosition(self, normalizedX, normalizedY, insetIndex);
end

-- copied
function TomCatsMapCanvasPinMixin:UseFrameLevelType(pinFrameLevelType, index)
	self.pinFrameLevelType = pinFrameLevelType;
	self.pinFrameLevelIndex = index;
end

-- copied
function TomCatsMapCanvasPinMixin:SetScaleStyle(scaleStyle)
	if scaleStyle == AM_PIN_SCALE_STYLE_VISIBLE_WHEN_ZOOMED_IN then
		self:SetScalingLimits(1.5, 0.0, 2.55);
	elseif scaleStyle == AM_PIN_SCALE_STYLE_VISIBLE_WHEN_ZOOMED_OUT then
		self:SetScalingLimits(1.5, 0.825, 0.0);
	elseif scaleStyle == AM_PIN_SCALE_STYLE_WITH_TERRAIN then
		self:SetScalingLimits(nil, nil, nil);
		if self:IsIgnoringGlobalPinScale() then
			self:SetScale(1.0);
		else
			self:SetScale(self:GetGlobalPinScale());
		end
	end
end

-- copied
function TomCatsMapCanvasPinMixin:ApplyCurrentScale()
	local scale;
	if self.startScale and self.startScale and self.endScale then
		local parentScaleFactor = 1.0 / self:GetMap():GetCanvasScale();
		scale = parentScaleFactor * Lerp(self.startScale, self.endScale, Saturate(self.scaleFactor * self:GetMap():GetCanvasZoomPercent()));
	elseif not self:IsIgnoringGlobalPinScale() then
		scale = 1;
	end
	if scale then
		if not self:IsIgnoringGlobalPinScale() then
			scale = scale * self:GetMap():GetGlobalPinScale();
		end
		self:SetScale(scale);
		self:ApplyCurrentPosition();
	end
end

-- copied
function TomCatsMapCanvasPinMixin:OnCanvasScaleChanged()
	self:ApplyCurrentScale();
	--self:ApplyCurrentAlpha();
end

-- copied
function TomCatsMapCanvasPinMixin:IsIgnoringGlobalPinScale()
	return not not self.ignoreGlobalPinScale;
end

-- copied
function TomCatsMapCanvasPinMixin:ApplyCurrentPosition()
	self:GetMap():ApplyPinPosition(self, self.normalizedX, self.normalizedY, self.insetIndex);
end
